# More Information: <https://cloud.google.com/source-repositories/docs/reference/rest>
from typing import Any

from gcp_pilot import exceptions
from gcp_pilot.base import DiscoveryMixin, GoogleCloudPilotAPI

RepoType = dict[str, Any]


class SourceRepository(DiscoveryMixin, GoogleCloudPilotAPI):
    _iam_roles = ["source.repos.create"]

    def __init__(self, **kwargs):
        super().__init__(
            serviceName="sourcerepo",
            version="v1",
            cache_discovery=False,
            **kwargs,
        )

    def _repo_path(self, repo: str, project_id: str | None = None) -> str:
        parent_path = self._project_path(project_id=project_id)
        return f"{parent_path}/repos/{repo}"

    def list_repos(self, project_id: str | None = None) -> RepoType:
        params = dict(
            name=self._project_path(project_id=project_id),
        )
        items = self._paginate(
            method=self.client.projects().repos().list,
            result_key="repos",
            params=params,
        )
        for item in items:
            yield item

    def get_repo(self, repo_name: str, project_id: str | None = None) -> RepoType:
        return self._execute(
            method=self.client.projects().repos().get,
            name=self._repo_path(repo=repo_name, project_id=project_id),
        )

    def create_repo(self, repo_name: str, project_id: str | None = None, exists_ok: bool = True) -> RepoType:
        parent = self._project_path(project_id=project_id)
        repo_path = self._repo_path(repo=repo_name, project_id=project_id)
        try:
            return self._execute(
                method=self.client.projects().repos().create,
                parent=parent,
                body={
                    "name": repo_path,
                },
            )
        except exceptions.AlreadyExists:
            if not exists_ok:
                raise
            return self.get_repo(repo_name=repo_name, project_id=project_id)


__all__ = ("SourceRepository",)



# More Information: https://developers.google.com/admin-sdk/directory/reference/rest
from typing import Any, Generator

from gcp_pilot.base import DiscoveryMixin, GoogleCloudPilotAPI

UserType = GroupType = MemberType = dict[str, Any]
USER_FIELDS = [
    "addresses",
    "addresses",
    "ageRanges",
    "biographies",
    "birthdays",
    "calendarUrls",
    "clientData",
    "coverPhotos",
    "emailAddresses",
    "events",
    "externalIds",
    "genders",
    "imClients",
    "interests",
    "locales",
    "locations",
    "memberships",
    "metadata",
    "miscKeywords",
    "names",
    "nicknames",
    "occupations",
    "organizations",
    "phoneNumbers",
    "photos",
    "relations",
    "sipAddresses",
    "skills",
    "urls",
    "userDefined",
]


class People(DiscoveryMixin, GoogleCloudPilotAPI):
    _scopes = [
        "https://www.googleapis.com/auth/contacts",
        "https://www.googleapis.com/auth/directory.readonly",
    ]

    def __init__(self, email: str, **kwargs):
        self.email = email
        super().__init__(
            serviceName="people",
            version="v1",
            cache_discovery=False,
            subject=email,
            **kwargs,
        )

    def _get_project_default_location(self, project_id: str | None = None) -> str | None:
        return None

    def get_people(self, query: str | None = None, fields: list[str] | None = None) -> Generator[UserType, None, None]:
        params = {
            "readMask": ",".join(fields or USER_FIELDS),
            "sources": ["DIRECTORY_SOURCE_TYPE_DOMAIN_PROFILE", "DIRECTORY_SOURCE_TYPE_DOMAIN_CONTACT"],
        }
        if query:
            params["query"] = query
            yield from self._list(
                method=self.client.people().searchDirectoryPeople,
                result_key="people",
                params=params,
            )
        else:
            yield from self._paginate(
                method=self.client.people().listDirectoryPeople,
                result_key="people",
                params=params,
            )


__all__ = ("People",)



# https://googleapis.dev/python/pubsub/latest/index.html
import base64
import json
from dataclasses import dataclass
from typing import Any, Callable, Generator

from google.api_core.exceptions import AlreadyExists, NotFound
from google.cloud import pubsub_v1
from google.protobuf.field_mask_pb2 import FieldMask  # pylint: disable=no-name-in-module
from google.pubsub_v1 import PushConfig, Subscription, Topic, types

from gcp_pilot.base import GoogleCloudPilotAPI


class CloudPublisher(GoogleCloudPilotAPI):
    _client_class = pubsub_v1.PublisherClient
    _service_name = "Cloud Pub/Sub"
    _google_managed_service = True

    def create_topic(
        self,
        topic_id: str,
        project_id: str | None = None,
        exists_ok: bool = True,
        labels: dict[str, str] | None = None,
    ) -> types.Topic:
        topic_path = self.client.topic_path(
            project=project_id or self.project_id,
            topic=topic_id,
        )
        topic_obj = types.Topic(
            name=topic_path,
            labels=labels,
        )
        try:
            topic = self.client.create_topic(
                request=topic_obj,
            )
        except AlreadyExists:
            if not exists_ok:
                raise
            topic = self.get_topic(topic_id=topic_id, project_id=project_id)
        return topic

    def update_topic(
        self,
        topic_id: str,
        project_id: str | None = None,
        labels: dict[str, str] | None = None,
    ) -> types.Topic:
        topic_path = self.client.topic_path(
            project=project_id or self.project_id,
            topic=topic_id,
        )
        topic_obj = types.Topic(
            name=topic_path,
            labels=labels,
        )
        return self.client.update_topic(
            request=types.UpdateTopicRequest(
                topic=topic_obj,
                update_mask=FieldMask(paths=["labels"]),
            ),
        )

    def get_topic(self, topic_id: str, project_id: str | None = None):
        topic_path = self.client.topic_path(
            project=project_id or self.project_id,
            topic=topic_id,
        )
        return self.client.get_topic(
            topic=topic_path,
        )

    def list_topics(
        self,
        prefix: str = "",
        suffix: str = "",
        project_id: str | None = None,
    ) -> Generator[Topic, None, None]:
        project_path = self._project_path(project_id=project_id)
        topics = self.client.list_topics(
            project=project_path,
        )
        for topic in topics:
            name = topic.name.split("/topics/")[-1]
            if name.startswith(prefix) and name.endswith(suffix):
                yield topic

    def publish(
        self,
        message: str,
        topic_id: str,
        project_id: str | None = None,
        attributes: dict[str, Any] | None = None,
    ) -> types.PublishResponse:
        topic_path = self.client.topic_path(
            project=project_id or self.project_id,
            topic=topic_id,
        )
        try:
            future = self.client.publish(
                topic=topic_path,
                data=message.encode(),
                **(attributes or {}),
            )
            return future.result()
        except NotFound:
            self.create_topic(
                topic_id=topic_id,
                project_id=project_id,
            )
            future = self.client.publish(
                topic=topic_path,
                data=message.encode(),
                **(attributes or {}),
            )
            return future.result()


class CloudSubscriber(GoogleCloudPilotAPI):
    _client_class = pubsub_v1.SubscriberClient
    _service_name = "Cloud Pub/Sub"
    _google_managed_service = True

    def list_subscriptions(
        self,
        prefix: str = "",
        suffix: str = "",
        project_id: str | None = None,
    ) -> Generator[Subscription, None, None]:
        all_subscriptions = self.client.list_subscriptions(
            project=f"projects/{project_id or self.project_id}",
        )
        for subscription in all_subscriptions:
            name = subscription.name.split("/subscriptions/")[-1]
            if name.startswith(prefix) and name.endswith(suffix):
                yield subscription

    def get_subscription(self, subscription_id: str, project_id: str | None = None) -> Subscription:
        subscription_path = self.client.subscription_path(
            project=project_id or self.project_id,
            subscription=subscription_id,
        )

        return self.client.get_subscription(
            subscription=subscription_path,
        )

    def delete_subscription(self, subscription_id: str, project_id: str | None = None) -> None:
        subscription_path = self.client.subscription_path(
            project=project_id or self.project_id,
            subscription=subscription_id,
        )

        return self.client.delete_subscription(
            subscription=subscription_path,
        )

    def create_subscription(
        self,
        topic_id: str,
        subscription_id: str,
        project_id: str | None = None,
        exists_ok: bool = True,
        auto_create_topic: bool = True,
        enable_message_ordering: bool = False,
        push_to_url: str | None = None,
        use_oidc_auth: bool = False,
    ) -> Subscription:
        topic_path = self.client.topic_path(
            project=project_id or self.project_id,
            topic=topic_id,
        )
        subscription_path = self.client.subscription_path(
            project=project_id or self.project_id,
            subscription=subscription_id,
        )

        push_config = None
        if push_to_url:
            push_config = PushConfig(
                push_endpoint=push_to_url,
                **(self.get_oidc_token(audience=push_to_url) if use_oidc_auth else {}),
            )

        subscription = Subscription(
            name=subscription_path,
            topic=topic_path,
            push_config=push_config,
            enable_message_ordering=enable_message_ordering,
        )

        try:
            return self.client.create_subscription(request=subscription)
        except NotFound:
            if not auto_create_topic:
                raise
            CloudPublisher().create_topic(
                topic_id=topic_id,
                project_id=project_id,
                exists_ok=False,
            )
            return self.client.create_subscription(request=subscription)
        except AlreadyExists:
            if not exists_ok:
                raise
            return self.get_subscription(subscription_id=subscription_id, project_id=project_id)

    def update_subscription(
        self,
        topic_id: str,
        subscription_id: str,
        project_id: str | None = None,
        push_to_url: str | None = None,
        use_oidc_auth: bool = False,
    ) -> Subscription:
        topic_path = self.client.topic_path(
            project=project_id or self.project_id,
            topic=topic_id,
        )
        subscription_path = self.client.subscription_path(
            project=project_id or self.project_id,
            subscription=subscription_id,
        )

        push_config = None
        if push_to_url:
            push_config = PushConfig(
                push_endpoint=push_to_url,
                **(self.get_oidc_token(audience=push_to_url) if use_oidc_auth else {}),
            )

        subscription = Subscription(
            name=subscription_path,
            topic=topic_path,
            push_config=push_config,
        )

        update_mask = {"paths": {"push_config"}}

        return self.client.update_subscription(request={"subscription": subscription, "update_mask": update_mask})

    def create_or_update_subscription(
        self,
        topic_id: str,
        subscription_id: str,
        project_id: str | None = None,
        auto_create_topic: bool = True,
        enable_message_ordering: bool = False,
        push_to_url: str | None = None,
        use_oidc_auth: bool = False,
    ) -> Subscription:
        try:
            return self.create_subscription(
                topic_id=topic_id,
                subscription_id=subscription_id,
                project_id=project_id,
                exists_ok=False,
                auto_create_topic=auto_create_topic,
                enable_message_ordering=enable_message_ordering,
                push_to_url=push_to_url,
                use_oidc_auth=use_oidc_auth,
            )
        except AlreadyExists:
            return self.update_subscription(
                topic_id=topic_id,
                subscription_id=subscription_id,
                project_id=project_id,
                push_to_url=push_to_url,
                use_oidc_auth=use_oidc_auth,
            )

    def subscribe(self, topic_id: str, subscription_id: str, callback: Callable, project_id: str | None = None):
        self.create_subscription(
            topic_id=topic_id,
            subscription_id=subscription_id,
            project_id=project_id,
        )

        subscription_path = self.client.subscription_path(
            project=project_id or self.project_id,
            subscription=subscription_id,
        )
        future = self.client.subscribe(
            subscription=subscription_path,
            callback=callback,
        )
        future.result()


@dataclass
class Message:
    id: str
    data: Any
    attributes: dict[str, Any]
    subscription: str

    @classmethod
    def load(cls, body: str | bytes | dict, parser: Callable = json.loads) -> "Message":
        # https://cloud.google.com/pubsub/docs/push#receiving_messages
        if isinstance(body, bytes):
            body = body.decode()
        if isinstance(body, str):
            body = json.loads(body)

        return Message(
            id=body["message"]["messageId"],
            attributes=body["message"]["attributes"],
            subscription=body["subscription"],
            data=parser(base64.b64decode(body["message"]["data"]).decode("utf-8")),
        )

    def dump(self, parser: Callable = json.dumps) -> str:
        body = {
            "message": {
                "messageId": self.id,
                "attributes": self.attributes,
                "data": base64.b64encode(parser(self.data).encode("utf-8")).decode(),
            },
            "subscription": self.subscription,
        }
        return json.dumps(body)


__all__ = (
    "CloudPublisher",
    "CloudSubscriber",
    "Message",
)



# More Information: https://cloud.google.com/api-keys/docs/reference/rest
from dataclasses import dataclass
from datetime import datetime
from typing import Generator

from gcp_pilot.base import DiscoveryMixin, GoogleCloudPilotAPI


@dataclass
class Key:
    raw: dict
    project_id: str

    @property
    def key_id(self) -> str:
        return self.raw["name"].rsplit("/", 1)[-1]

    @property
    def uid(self) -> str:
        return self.raw["uid"]

    @property
    def etag(self) -> str:
        return self.raw["etag"]

    @property
    def api_targets(self) -> list:
        return self.raw["restrictions"].get("apiTargets", [])

    @property
    def display_name(self) -> str:
        return self.raw["displayName"]

    @property
    def created_at(self) -> datetime:
        return datetime.strptime(self.raw["createTime"], "%Y-%m-%dT%H:%M:%S.%fZ")

    @property
    def updated_at(self) -> datetime:
        return datetime.strptime(self.raw["updateTime"], "%Y-%m-%dT%H:%M:%S.%fZ")

    @property
    def value(self) -> str:
        data = APIKey().get_key_string(key_id=self.key_id)
        return data["keyString"]


class APIKey(DiscoveryMixin, GoogleCloudPilotAPI):
    def __init__(self, **kwargs):
        super().__init__(
            serviceName="apikeys",
            version="v2",
            location="global",
            **kwargs,
        )

    def _key_path(self, key_id: str, project_id: str | None = None) -> str:
        location_path = self._location_path(project_id=project_id)
        return f"{location_path}/keys/{key_id}"

    def lookup(self, key: str):
        return self._execute(
            method=self.client.keys().lookupKey,
            parent=key,
        )

    def get(self, key_id: str, project_id: str | None = None) -> Key:
        data = self._execute(
            method=self.client.projects().locations().keys().get,
            name=self._key_path(key_id=key_id, project_id=project_id),
        )
        return Key(raw=data, project_id=project_id or self.project_id)

    def create(
        self,
        key_id: str,
        display_name: str = "",
        api_targets: list[str] | None = None,
        project_id: str | None = None,
    ) -> dict:
        body = {
            "displayName": display_name,
            "restrictions": {},
        }
        if api_targets:
            body["restrictions"]["apiTargets"] = [{"service": service} for service in api_targets]

        return self._execute(
            method=self.client.projects().locations().keys().create,
            parent=self._location_path(project_id=project_id),
            keyId=key_id,
            body=body,
        )

    def delete(self, key_id: str, project_id: str | None = None):
        return self._execute(
            method=self.client.projects().locations().keys().delete,
            name=self._key_path(key_id=key_id, project_id=project_id),
        )

    def undelete(self, key_id: str, project_id: str | None = None):
        return self._execute(
            method=self.client.projects().locations().keys().undelete,
            name=self._key_path(key_id=key_id, project_id=project_id),
        )

    def list(self, project_id: str | None = None) -> Generator[Key, None, None]:
        params = dict(parent=self._location_path(project_id=project_id))
        data = self._paginate(
            method=self.client.projects().locations().keys().list,
            result_key="keys",
            params=params,
        )
        for item in data:
            yield Key(raw=item, project_id=project_id or self.project_id)

    def get_key_string(self, key_id: str, project_id: str | None = None) -> dict:
        return self._execute(
            method=self.client.projects().locations().keys().getKeyString,
            name=self._key_path(key_id=key_id, project_id=project_id),
        )
